from __future__ import annotations

import asyncio
from typing import Any, AsyncIterable, Callable, Coroutine, Iterable, cast

from synkka.exceptions import TaskFailed
from synkka.types import TIn, TOut
from synkka.utilities import async_chunked, iterable_to_async_iterable


class Sentinel:
    pass


async def imap_chunked_pairs(
    func: Callable[[TIn], Coroutine[Any, Any, TOut]],
    items: Iterable[TIn] | AsyncIterable[TIn],
    chunk_size: int,
) -> AsyncIterable[tuple[TIn, TOut]]:
    """
    Maps `func` over `items` in chunks of size `chunk_size`.
    Yields pairs of (item, result).
    """
    chunk: list[TIn]
    async for chunk in async_chunked(items, chunk_size):
        for pair in zip(chunk, await asyncio.gather(*[func(thing) for thing in chunk])):
            yield pair


async def imap_unordered_pairs(
    func: Callable[[TIn], Coroutine[Any, Any, TOut]],
    items: Iterable[TIn] | AsyncIterable[TIn],
    concurrency: int,
) -> AsyncIterable[tuple[TIn, TOut]]:
    """
    Maps `func` over `items` with `concurrency` workers.
    Yields pairs of (item, result).
    """
    sentinel = Sentinel()
    inbox: asyncio.Queue[TIn | Sentinel] = asyncio.Queue()
    outbox: asyncio.Queue[
        tuple[TIn | Sentinel, TOut | None, BaseException | None]
    ] = asyncio.Queue()
    cancel = False
    n_workers_quit = 0

    async def worker():
        nonlocal n_workers_quit
        while not cancel:
            work_item = await inbox.get()
            if work_item is sentinel:
                break
            try:
                work_result = await func(work_item)
            except Exception as e:
                await outbox.put((work_item, None, e))
            else:
                await outbox.put((work_item, work_result, None))
        n_workers_quit += 1
        if n_workers_quit == concurrency:
            await outbox.put((sentinel, None, None))

    async for item in iterable_to_async_iterable(items):
        inbox.put_nowait(item)
    for _ in range(concurrency):
        inbox.put_nowait(sentinel)

    workers = [asyncio.create_task(worker()) for _ in range(concurrency)]
    while True:
        res_item, result, exception = await outbox.get()
        if res_item is sentinel:
            break
        assert not isinstance(res_item, Sentinel)
        if exception is not None:
            cancel = True
            raise TaskFailed(str(exception), item=res_item) from exception
        yield (res_item, cast(TOut, result))
    await asyncio.gather(*workers)



from __future__ import annotations

from typing import AsyncIterable, Iterable, cast

from synkka.types import TIn


def iterable_to_async_iterable(
    iterable: Iterable[TIn] | AsyncIterable[TIn],
) -> AsyncIterable[TIn]:
    """
    Converts an iterable to an async iterable.
    """
    if hasattr(iterable, "__aiter__"):
        return cast(AsyncIterable[TIn], iterable)

    async def async_iter():
        for item in iterable:
            yield item

    return async_iter()


async def async_chunked(
    items: Iterable[TIn] | AsyncIterable[TIn], chunk_size: int
) -> AsyncIterable[list[TIn]]:
    """
    Yields chunks of `items` of (up to) size `chunk_size`.
    """
    chunk: list[TIn] = []
    async for item in iterable_to_async_iterable(items):
        chunk.append(item)
        if len(chunk) == chunk_size:
            yield chunk
            chunk = []
    if chunk:
        yield chunk



import asyncio

import pytest

from synkka.exceptions import TaskFailed
from synkka.iterators import imap_chunked_pairs, imap_unordered_pairs


@pytest.mark.asyncio
@pytest.mark.parametrize("imap_fun", [imap_chunked_pairs, imap_unordered_pairs])
async def test_imap_pairs_functions(imap_fun):
    in_flight = 0
    flight_limit = 5

    async def do_thing(value):
        nonlocal in_flight
        try:
            in_flight += 1
            await asyncio.sleep(0)
            return value * 10
        finally:
            assert in_flight <= flight_limit, "Too many things in flight"
            in_flight -= 1

    async for a, b in imap_fun(do_thing, range(50), flight_limit):
        assert a * 10 == b


@pytest.mark.asyncio
async def test_imap_unordered_pairs_does_no_work_after_first_exception():
    did_fail = False

    async def do_thing(value):
        nonlocal did_fail
        assert not did_fail  # We shouldn't enter this function after the first failure
        if value == 5:
            did_fail = True
            raise ValueError("Nope")
        await asyncio.sleep(0)
        return value * 10

    with pytest.raises(TaskFailed) as ei:
        async for a, b in imap_unordered_pairs(do_thing, range(10), 2):
            pass
    assert ei.value.args[0] == "Nope"
    assert isinstance(ei.value.__cause__, ValueError)  # original exception is preserved
    assert did_fail
